// Complex functions

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>

#define COMPLEX double complex

#if 0	// math library built-in function

COMPLEX Sin(x) COMPLEX x;
{
  double e, f;

  e = exp(cimag(x)); f = 1.0 / e;
  x = 0.5 * cos(creal(x)) * (e - f) + 0.5 * sin(creal(x)) * (e + f) * I;
  return x;
}

COMPLEX Cos(x) COMPLEX x;
{
  double e, f;

  e = exp(cimag(x)); f = 1.0 / e;
  x = 0.5 * cos(creal(x)) * (f + e) + 0.5 * sin(creal(x)) * (f - e) * I;
  return x;
}

COMPLEX Tan(x) COMPLEX x;
{
  double e, f, d;

  e = exp(2.0 * cimag(x)); f = 1.0 / e;
  d = cos(2.0 * creal(x)) + 0.5 * (e + f);
  x = sin(2.0 * creal(x)) / d + 0.5 * (e - f) / d * I;
  return x;
}

COMPLEX Sinh(x) COMPLEX x;
{
  double e, f;

  e = exp(creal(x)); f = 1.0 / e;
  x = 0.5 * (e - f) * cos(cimag(x)) + 0.5 * (e + f) * sin(cimag(x)) * I;
  return x;
};

COMPLEX Cosh(x) COMPLEX x;
{
  double e, f;

  e = exp(creal(x)); f = 1.0 / e;
  x = 0.5 * (e + f) * cos(cimag(x)) + 0.5 * (e - f) * sin(cimag(x)) * I;
  return x;
};

COMPLEX Tanh(x) COMPLEX x;
{
  double e, f, d;

//return Sinh(x) / Cosh(x);
  e = exp(2.0 * creal(x)); f = 1.0 / e;
  d = 0.5 * (e + f) + cos(2.0 * cimag(x));
  x = 0.5 * (e - f) / d + sin(2.0 * cimag(x)) / d * I;
  return x;
};

COMPLEX Exp(x) COMPLEX x;
{
  double a;

  a = exp(creal(x));
  x = a * cos(cimag(x)) + a * sin(cimag(x)) * I;
  return x;
}
#endif

COMPLEX Log(x) COMPLEX x;
{
  COMPLEX z;

  if (x == (COMPLEX)0.0)
        return -INFINITY;
  z = 0.5 * log(creal(x) * creal(x) + cimag(x) * cimag(x))
        +  atan2(cimag(x), creal(x)) * I;
  return z;
}

COMPLEX Asinh(x) COMPLEX x;
{
  return Log(x + csqrt(x * x + 1));
}

COMPLEX Acosh(x) COMPLEX x;
{
  return Log(x + csqrt(x * x - 1));
}

COMPLEX Atanh(x) COMPLEX x;
{
  return 0.5 * Log((1.0 + x) / (1.0 - x));
}


COMPLEX Pow(x, y) COMPLEX x, y;
{
  if (x == (COMPLEX)0.0 && y == (COMPLEX)0.0)
        return NAN;
  else if (y == (COMPLEX)0.0)
        return (COMPLEX)1.0;
  else if (x == (COMPLEX)0.0)
        return (COMPLEX)0.0;
  else if (y == (COMPLEX)1.0)
        return x;
  return cexp(y * Log(x));
}

/*
 * Complex FFT
 * inv = 0: fft, inv = 1: ifft
 */

void FFT(COMPLEX *z, int n, int inv)
{
  int l, nv2, ip, i, j, k, m, le, le1;
  COMPLEX t, w, u;
  double c, s, f;

  f = inv ? 1.0 : -1.0;
  l = (int)(log((double)n) / log(2.0) + 0.001);
  if ((int)pow(2.0, (double)l) != n) {
	fprintf(stderr, "FFT(): n(%d) must be a power of 2.\n", n);
	exit(1);
  }
  nv2 = n / 2;
  for (i = j = 0; i < n - 1; i++) {	/* bit reverse */
	if (i <= j) {
		t = z[i]; z[i] = z[j]; z[j] = t;
	}
	k = nv2;
	while (k <= j) {
		j -= k; k /= 2;
	}
	j += k;
  }
  for (m = 1; m <= l; m++) {
	le = (int)(pow(2.0, m) + 0.001);
	le1 = le / 2;
	u = 1.0 + 0.0 * I;
	w = cos(M_PI / (double)le1) + f * sin(M_PI / (double)le1) * I;
	for (j = 0; j < le / 2; j++) {
		for (i = j; i < n; i += le) {
			ip = i + le1;
			t = z[ip] * u;
			z[ip] = z[i] - t;
			z[i] = z[i] + t;
		}
		u *= w;
	}
  }
}

// Imput impedance of cable
// Zt = terminal load (Ohm)
// Z0 = cable characterristic impedance (Ohm)
// Vr = cable velocity ratio (0 < Vr <= 1)
// alpha = cable attenuation constant (neper/m)
// len = cable length
// f = frequenct (Hz)

COMPLEX Zin(Zt, Z0, Vr, alpha, len, f) COMPLEX Zt; double Z0, Vr, alpha, len, f;
{
  double omega, beta;
  COMPLEX gamma, z1;

  omega = 2.0 * M_PI * f;
  beta = omega / (2.99792458e8 * Vr) + alpha;
  gamma = alpha + beta * I;
  z1 = ctanh(gamma * len);
  return Z0 * (Zt / Z0 + z1)
	/ (1.0 + (Zt / Z0) * z1);
};

// Cable transfer function: V2 / V1
// V1 = cable input voltage (V)
// V2 = cable output voltage (V)
// Z0 = cable characteristic impedance (real part only)
// Vr = cable velocity rartio (0 < Vr <= 1)
// alpha = cable attenuation constant (neper/m)
// len = cable length (m)
// Zs = transmiter impedance (Ohm)
// Tr = receiver impedance (Ohm)

COMPLEX Trans(Zs, Zr, Z0, Vr, alpha, len, f) COMPLEX Zs, Zr;
	double Z0, Vr, alpha, len, f;
{
  double omega, beta;
  COMPLEX gamma, Zi;

  omega = 2.0 * M_PI * f;
  beta = omega / (2.99792458e8 * Vr) + alpha;
  gamma = alpha + beta * I;
  Zi = Zin(Zs, Z0, Vr, alpha, len, f);
  return ccosh(gamma * len) - Z0 * csinh(gamma * len) / Zi;
}

// Cable S-parameter
// Zs = system impedance (usually 50 + j*0 Ohm)
// Z0 = cable characteristic impedance (real part only)
// Vr = cable velocity rartio (0 < Vr <= 1)
// alpha = cable attenuation constant (neper/m)
// len = cable length (m)

COMPLEX S11(Zs, Z0, Vr, alpha, len, f) double Zs, Z0, Vr, alpha, len, f;
{
#if 0
  COMPLEX Zi, Zt, S;

  Zt = Zs + 0.0 * I;
  Zi = Zin(Zt, Z0, Vr, alpha, len, f);
  S = (Zi - Zs) / (Zi + Zs);
#else
  double c, omega, beta;
  COMPLEX gamma, S;

  omega = 2.0 * M_PI * f;
  beta = omega / (2.99792458e8 * Vr) + alpha;
  gamma = alpha + beta * I;

  S = (Z0 / Zs - Zs / Z0) * ctanh(gamma * len)
	/ (2.0 +(Z0 / Zs + Zs / Z0) * ctanh(gamma * len));
#endif
  if (creal(S) < 1e-9)
	S = 0.0 + 0.0 * I;
  return S;
}

COMPLEX S21(Zs, Z0, Vr, alpha, len, f) COMPLEX Zs;
	double Z0, Vr, alpha, len, f;
{
  double c, omega, beta;
  COMPLEX gamma, Zi, V2;

  omega = 2.0 * M_PI * f;
  beta = omega / (2.99792458e8 * Vr) + alpha;
  gamma = alpha + beta * I;
#if 0
  Zi = Zin(Zs, Z0, Vr, alpha, len, f);
  V2 = Cosh(gamma * len) - Z0 * Sinh(gamma * len) / Zi;
//  return 2.0 * V2 / (1.0 + Zs / Zi);
  return 2.0 * V2 * Zi / (Zi + Zs);
#else
  return 2.0 / (2.0 * ccosh(gamma * len)
	+ (Z0 / Zs + Zs / Z0) * csinh(gamma * len));
#endif
};

// Find Z0 from the S-parameter

COMPLEX S2Z(S11, S21, Zs) COMPLEX S11, S21; double Zs;
{
  COMPLEX Z, Z1, Z2;

  Z1 = 1 + S11;
  Z2 = 1 - S11;
  Z = (Z1 * Z1 - S21 * S21) / (Z2 * Z2 - S21 * S21) * Zs * Zs;
  return csqrt(Z);
}

// Find Gamma from the S-parameter

COMPLEX S2G(S11, S21) COMPLEX S11, S21;
{
  COMPLEX Z, Z1, Z2;

  Z1 = 1.0 + S11;
  Z2 = 1.0 - S11;
  Z = (Z1 * Z1 - S21 * S21) * (Z2 * Z2 - S21 * S21) / 4.0 / S21 / S21;
//  Z = (Z1 * Z1 / S21 - S21) * (Z2 * Z2 / S21 - S21) / 4.0;
  return Asinh(csqrt(Z));
};


// test

COMPLEX Gy(Zs, Z0, Vr, alpha, len, f) COMPLEX Zs;
	double Z0, Vr, alpha, len, f;
{
  double c, omega, beta;
  COMPLEX gamma, Zi, V2;

  omega = 2.0 * M_PI * f;
  beta = omega / (2.99792458e8 * Vr) + alpha;
  gamma = alpha + beta * I;
  return gamma * len;
}
